# -*- coding: utf-8 -*-
#importing modules
from __future__ import absolute_import, division, print_function, unicode_literals
from .colorprint import colorprint, OKGREEN, FAIL
from .threadpool import ThreadPool
from collections import defaultdict
import ssl
import os
import sys

IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    # Python 2
    from urllib2 import Request, urlopen, HTTPError, URLError
else:
    # Python 3
    from urllib.request import Request, urlopen, HTTPError, URLError

    unicode = str

#global variable
MAX_THREADS_DEFAULT = 7

# Used to allow downloading files even if https certificate doesn't match
if hasattr(ssl, "_create_unverified_context"):
    ssl_unverified_context = ssl._create_unverified_context()
else:
    # Not existing in Python 2.6
    ssl_unverified_context = None


def sanitize_url(url):
    """ Make sure this url works with urllib2 (ascii, http, etc) """
    if url and not url.startswith("http"):
        url = "http://%s" % url
    url = url.encode("ascii", "ignore").decode("utf-8")
    return url


def get_status_code(url):
    """ Perform HEAD request and return status code """
    try:
        request = Request(sanitize_url(url))
        request.add_header(
            "User-Agent",
            "Mozilla/5.0 (compatible; MSIE 9.0; " "Windows NT 6.1; Trident/5.0)",
        )
        request.get_method = lambda: "HEAD"
        response = urlopen(request, context=ssl_unverified_context)
        # print response.info()
        return response.getcode()
    except HTTPError as e:
        return e.code
    except URLError as e:
        return e.reason
    except Exception as e:
        print(e, url)
        return None


def check_refs(refs, verbose=True, max_threads=MAX_THREADS_DEFAULT):
    """ Check if urls exist """
    codes = defaultdict(list)

    def check_url(ref):
        url = ref.ref
        status_code = str(get_status_code(url))
        codes[status_code].append(ref)
        if verbose:
            if status_code == "200":
                colorprint(OKGREEN, "%s - %s" % (status_code, url))
            else:
                colorprint(FAIL, "%s - %s" % (status_code, url))

    # Start a threadpool and add the check-url tasks
    try:
        pool = ThreadPool(5)
        pool.map(check_url, refs)
        pool.wait_completion()

    except Exception as e:
        print(e)
    except KeyboardInterrupt:
        pass

    # Print summary
    print("\nSummary of link checker:")
    if "200" in codes:
        colorprint(OKGREEN, "%s working" % len(codes["200"]))
    for c in sorted(codes):
        if c != "200":
            colorprint(FAIL, "%s broken (reason: %s)" % (len(codes[c]), c))
            for ref in codes[c]:
                o = "  - %s" % ref.ref
                if ref.page > 0:
                    o += " (page %s)" % ref.page
                print(o)


def download_urls(
    urls, output_directory, verbose=True, max_threads=MAX_THREADS_DEFAULT
):
    """ Download urls to a target directory """
    assert type(urls) in [list, tuple, set], "Urls must be some kind of list"
    assert len(urls), "Need urls"
    assert output_directory, "Need an output_directory"

    def vprint(s):
        if verbose:
            print(s)
#Download URL content in Directory
    def download_url(url):
        try:
            fn = url.split("/")[-1].split("?")[0]
            fn_download = os.path.join(output_directory, fn)
            with open(fn_download, "wb") as f:
                request = Request(sanitize_url(url))
                request.add_header(
                    "User-Agent",
                    "Mozilla/5.0 (compatible; "
                    "MSIE 9.0; Windows NT 6.1; Trident/5.0)",
                )
                response = urlopen(request, context=ssl_unverified_context)
                status_code = response.getcode()
                if status_code == 200:
                    f.write(urlopen(request).read())
                    colorprint(OKGREEN, "Downloaded '%s' to '%s'" % (url, fn_download))
                else:
                    colorprint(FAIL, "Error downloading '%s' (%s)" % (url, status_code))
        except HTTPError as e:
            colorprint(FAIL, "Error downloading '%s' (%s)" % (url, e.code))
        except URLError as e:
            colorprint(FAIL, "Error downloading '%s' (%s)" % (url, e.reason))
        except Exception as e:
            colorprint(FAIL, "Error downloading '%s' (%s)" % (url, str(e)))

    # Create directory
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
        vprint("Created directory '%s'" % output_directory)
#Start a threadpool and add the download url tasks.
    try:
        pool = ThreadPool(5)
        pool.map(download_url, urls)
        pool.wait_completion()

    except Exception as e:
        print(e)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Command line tool to get metadata and URLs from a local or remote PDF,
and optionally download all referenced PDFs.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import argparse
import json
import codecs

import linkrot
from linkrot.downloader import check_refs


IS_PY2 = sys.version_info < (3, 0)
if IS_PY2:
    # Python 2
    parse_str = unicode  # noqa: F821
else:
    # Python 3
    parse_str = str

# print(sys.version)
# print("stdout encoding: %s" % sys.stdout.encoding)


def exit_with_error(code, *objs):
    print("ERROR %s:" % code, *objs, file=sys.stderr)
    exit(code)


# Error Status Codes
ERROR_FILE_NOT_FOUND = 1
ERROR_DOWNLOAD = 2
ERROR_PDF_INVALID = 4


def create_parser():
    parser = argparse.ArgumentParser(
        description="Extract metadata and references from a PDF, and "
        "optionally download all referenced PDFs. Visit "
        "https://github.com/marshalmiller/linkrot for more information.",
        epilog="",
    )

    parser.add_argument("pdf", help="Filename or URL of a PDF file")

    parser.add_argument(
        "-d",
        "--download-pdfs",
        metavar="OUTPUT_DIRECTORY",
        help="Download all referenced PDFs into specified directory",
    )

    parser.add_argument(
        "-c", "--check-links", action="store_true", help="Check for broken links"
    )

    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="Output infos as JSON (instead of plain text)",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Print all references (instead of only PDFs)",
    )

    # parser.add_argument("--debug",
    #                     action='store_true',
    #                     help="Output debug infos")

    parser.add_argument(
        "-t",
        "--text",
        action="store_true",
        help="Only extract text (no metadata or references)",
    )

    parser.add_argument(
        "-o", "--output-file", help="Output to specified file instead of console"
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s v{version}".format(version=linkrot.__version__),
    )
    return parser


def get_text_output(pdf, args):
    """ Normal output of infos of linkrot instance """
    # Metadata
    ret = ""
    ret += "Document infos:\n"
    for k, v in sorted(pdf.get_metadata().items()):
        if v:
            ret += "- %s = %s\n" % (k, parse_str(v).strip("/"))

    # References
    ref_cnt = pdf.get_references_count()
    ret += "\nReferences: %s\n" % ref_cnt
    refs = pdf.get_references_as_dict()
    for k in refs:
        ret += "- %s: %s\n" % (k.upper(), len(refs[k]))

    if args.verbose == 0:
        if "pdf" in refs:
            ret += "\nPDF References:\n"
            for ref in refs["pdf"]:
                ret += "- %s\n" % ref
        elif ref_cnt:
            ret += "\nTip: You can use the '-v' flag to see all references\n"
    else:
        if ref_cnt:
            for reftype in refs:
                ret += "\n%s References:\n" % reftype.upper()
                for ref in refs[reftype]:
                    ret += "- %s\n" % ref

    return ret.strip()


def print_to_console(text):
    # Prints a (unicode) string to the console, encoded depending on the stdout
    # encoding (eg. cp437 on Windows). Works with Python 2 and 3.
    try:
        sys.stdout.write(text)
    except UnicodeEncodeError:
        bytes_string = text.encode(sys.stdout.encoding, "backslashreplace")
        if hasattr(sys.stdout, "buffer"):
            sys.stdout.buffer.write(bytes_string)
        else:
            text = bytes_string.decode(sys.stdout.encoding, "strict")
            sys.stdout.write(text)
    sys.stdout.write("\n")


def main():
    parser = create_parser()
    args = parser.parse_args()

    # if args.debug:
    #     logging.basicConfig(
    #             level=logging.DEBUG,
    #             format='%(levelname)s - %(module)s - %(message)s')

    try:
        pdf = linkrot.linkrot(args.pdf)
    except linkrot.exceptions.FileNotFoundError as e:
        exit_with_error(ERROR_FILE_NOT_FOUND, str(e))
    except linkrot.exceptions.DownloadError as e:
        exit_with_error(ERROR_DOWNLOAD, str(e))
    except linkrot.exceptions.PDFInvalidError as e:
        exit_with_error(ERROR_PDF_INVALID, str(e))

    # Perhaps only output text
    if args.text:
        text = pdf.get_text()
        if args.output_file:
            # to file (in utf-8)
            with codecs.open(args.output_file, "w", "utf-8") as f:
                f.write(text)
        else:
            # to console
            print_to_console(text)
        return

    # Print Metadata
    if args.json:
        # in JSON format
        text = json.dumps(pdf.summary, indent=4)
        if args.output_file:
            # to file (in utf-8)
            with codecs.open(args.output_file, "w", "utf-8") as f:
                f.write(text)
        else:
            # to console
            print_to_console(text)
    else:
        # in text format
        text = get_text_output(pdf, args)
        if args.output_file:
            # to file (in utf-8)
            with codecs.open(args.output_file, "w", "utf-8") as f:
                f.write(text)
        else:
            # to console
            print_to_console(text)

    if args.check_links:
        refs_all = pdf.get_references()
        refs = [ref for ref in refs_all if ref.reftype in ["url", "pdf"]]
        print("\nChecking %s URLs for broken links..." % len(refs))
        check_refs(refs)

    try:
        if args.download_pdfs:
            print(
                "\nDownloading %s pdfs to '%s'..."
                % (len(pdf.get_references("pdf")), args.download_pdfs)
            )
            pdf.download_pdfs(args.download_pdfs)
            print("All done!")
    except Exception as e:
        exit_with_error(ERROR_DOWNLOAD, str(e))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Extract metadata and links from a local or remote PDF, and
optionally download all referenced PDFs.

Features

* Extract metadata and PDF URLs from a given PDF
* Download all PDFs referenced in the original PDF
* Works with local and online pdfs
* Use as command-line tool or Python package
* Compatible with Python 2 and 3

Usage

linkrot can be used to extract infos from PDF in two ways:

* Command line tool `linkrot`
* Python library `import linkrot`

>>> import linkrot
>>> pdf = linkrot.linkrot("filename-or-url.pdf")
>>> metadata = pdf.get_metadata()
>>> references_list = pdf.get_references()
>>> references_dict = pdf.get_references_as_dict()
>>> pdf.download_pdfs("target-directory")

https://github.com/marshalmiller/linkrot

Copyright (c) 2021, Marshal Miller <marshal@marshalmiller.com>
License: GPLv3
"""
from __future__ import absolute_import, division, print_function, unicode_literals

__title__ = "linkrot"
__version__ = "1.2"
__author__ = "Marshal Miller"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Marshal Miller"

import os
import sys
import json
import shutil
import logging


from .extractor import extract_urls
from .backends import PDFMinerBackend, TextBackend
from .downloader import download_urls
from .exceptions import FileNotFoundError, DownloadError, PDFInvalidError
from pdfminer.pdfparser import PDFSyntaxError


IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    # Python 2
    from cStringIO import StringIO as BytesIO
    from urllib2 import Request, urlopen
else:
    # Python 3
    from io import BytesIO
    from urllib.request import Request, urlopen

    unicode = str

logger = logging.getLogger(__name__)


class linkrot(object):
    """
    Main class which extracts infos from PDF

    General flow:
    * init -> get_metadata()

    In detail:
    >>> import linkrot
    >>> pdf = linkrot.linkrot("filename-or-url.pdf")
    >>> print(pdf.get_metadata())
    >>> print(pdf.get_tet())
    >>> print(pdf.get_references())
    >>> pdf.download_pdfs("target-directory")
    """

    # Available after init
    uri = None  # Original URI
    fn = None  # Filename part of URI
    is_url = False  # False if file
    is_pdf = True

    stream = None  # ByteIO Stream
    reader = None  # ReaderBackend
    summary = {}

    def __init__(self, uri):
        """
        Open PDF handle and parse PDF metadata
        - `uri` can bei either a filename or an url
        """
        logger.debug("Init with uri: %s" % uri)

        self.uri = uri

        # Find out whether pdf is an URL or local file
        url = extract_urls(uri)
        self.is_url = len(url)

        # Grab content of reference
        if self.is_url:
            logger.debug("Reading url '%s'..." % uri)
            self.fn = uri.split("/")[-1]
            try:
                content = urlopen(Request(uri)).read()
                self.stream = BytesIO(content)
            except Exception as e:
                raise DownloadError("Error downloading '%s' (%s)" % (uri, unicode(e)))

        else:
            if not os.path.isfile(uri):
                raise FileNotFoundError("Invalid filename and not an url: '%s'" % uri)
            self.fn = os.path.basename(uri)
            self.stream = open(uri, "rb")

        # Create ReaderBackend instance
        try:
            self.reader = PDFMinerBackend(self.stream)
        except PDFSyntaxError as e:
            raise PDFInvalidError("Invalid PDF (%s)" % unicode(e))

            # Could try to create a TextReader
            logger.info(unicode(e))
            logger.info("Trying to create a TextReader backend...")
            self.stream.seek(0)
            self.reader = TextBackend(self.stream)
            self.is_pdf = False
        except Exception as e:
            raise
            raise PDFInvalidError("Invalid PDF (%s)" % unicode(e))

        # Save metadata to user-supplied directory
        self.summary = {
            "source": {
                "type": "url" if self.is_url else "file",
                "location": self.uri,
                "filename": self.fn,
            },
            "metadata": self.reader.get_metadata(),
        }

        # Search for URLs
        self.summary["references"] = self.reader.get_references_as_dict()
        # print(self.summary)

    def get_text(self):
        return self.reader.get_text()

    def get_metadata(self):
        return self.reader.get_metadata()

    def get_references(self, reftype=None, sort=False):
        """ reftype can be `None` for all, `pdf`, etc. """
        return self.reader.get_references(reftype=reftype, sort=sort)

    def get_references_as_dict(self, reftype=None, sort=False):
        """ reftype can be `None` for all, `pdf`, etc. """
        return self.reader.get_references_as_dict(reftype=reftype, sort=sort)

    def get_references_count(self, reftype=None):
        """ reftype can be `None` for all, `pdf`, etc. """
        r = self.reader.get_references(reftype=reftype)
        return len(r)

    def download_pdfs(self, target_dir):
        logger.debug("Download pdfs to %s" % target_dir)
        assert target_dir, "Need a download directory"
        assert not os.path.isfile(target_dir), "Download directory is a file"

        # Create output directory
        if target_dir and not os.path.exists(target_dir):
            os.makedirs(target_dir)
            logger.debug("Created output directory '%s'" % target_dir)

        # Save original PDF to user-supplied directory
        fn = os.path.join(target_dir, self.fn)
        with open(fn, "wb") as f:
            self.stream.seek(0)
            shutil.copyfileobj(self.stream, f)
        logger.debug("- Saved original pdf as '%s'" % fn)

        fn_json = "%s.infos.json" % fn
        with open(fn_json, "w") as f:
            f.write(json.dumps(self.summary, indent=2))
        logger.debug("- Saved metadata to '%s'" % fn_json)

        # Download references
        urls = [ref.ref for ref in self.get_references("pdf")]
        if not urls:
            return

        dir_referenced_pdfs = os.path.join(target_dir, "%s-referenced-pdfs" % self.fn)
        logger.debug("Downloading %s referenced pdfs..." % len(urls))

        # Download urls as a set to avoid duplicates
        download_urls(urls, dir_referenced_pdfs)



# -*- coding: utf-8 -*-
"""
PDF Backend: pdfMiner
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import logging
from io import BytesIO
from re import compile

# Character Detection Helper
import chardet

# Find URLs in text via regex
from . import extractor
from .libs.xmp import xmp_to_dict

# Setting `psparser.STRICT` is the first thing to do because it is
# referenced in the other pdfparser modules
from pdfminer import settings as pdfminer_settings

pdfminer_settings.STRICT = False
from pdfminer import psparser  # noqa: E402
from pdfminer.pdfdocument import PDFDocument  # noqa: E402
from pdfminer.pdfparser import PDFParser  # noqa: E402
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter  # noqa: E402
from pdfminer.pdfpage import PDFPage  # noqa: E402
from pdfminer.pdftypes import resolve1, PDFObjRef  # noqa: E402
from pdfminer.converter import TextConverter  # noqa: E402
from pdfminer.layout import LAParams  # noqa: E402


logger = logging.getLogger(__name__)


IS_PY2 = sys.version_info < (3, 0)
if not IS_PY2:
    # Python 3
    unicode = str


def make_compat_str(in_str):
    """
    Tries to guess encoding of [str/bytes] and
    return a standard unicode string
    """
    assert isinstance(in_str, (bytes, str, unicode))
    if not in_str:
        return unicode()

    # Chardet in Py2 works on str + bytes objects
    if IS_PY2 and isinstance(in_str, unicode):
        return in_str

    # Chardet in Py3 works on bytes objects
    if not IS_PY2 and not isinstance(in_str, bytes):
        return in_str

    # Detect the encoding now
    enc = chardet.detect(in_str)

    # Decode the object into a unicode object
    out_str = in_str.decode(enc["encoding"])

    # Cleanup
    if enc["encoding"] == "UTF-16BE":
        # Remove byte order marks (BOM)
        if out_str.startswith("\ufeff"):
            out_str = out_str[1:]
    return out_str


class Reference(object):
    """ Generic Reference """

    ref = ""
    reftype = "url"
    page = 0

    def __init__(self, uri, page=0):
        self.ref = uri
        self.reftype = "url"
        self.page = page

        self.pdf_regex = compile(r"\.pdf(:?\?.*)?$")

        # Detect reftype by filetype
        if self.pdf_regex.search(uri.lower()):
            self.reftype = "pdf"
            return

        # Detect reftype by extractor
        arxiv = extractor.extract_arxiv(uri)
        if arxiv:
            self.ref = arxiv.pop()
            self.reftype = "arxiv"
            return

        doi = extractor.extract_doi(uri)
        if doi:
            self.ref = doi.pop()
            self.reftype = "doi"
            return

    def __hash__(self):
        return hash(self.ref)

    def __eq__(self, other):
        assert isinstance(other, Reference)
        return self.ref == other.ref

    def __str__(self):
        return "<%s: %s>" % (self.reftype, self.ref)


class ReaderBackend(object):
    """
    Base class of all Readers (eg. for PDF files, text, etc.)

    The job of a Reader is to extract Text and Links.
    """

    text = ""
    metadata = {}
    references = set()

    def __init__(self):
        self.text = ""
        self.metadata = {}
        self.references = set()

    def get_metadata(self):
        return self.metadata

    def metadata_key_cleanup(self, d, k):
        """ Recursively clean metadata dictionaries """
        if isinstance(d[k], (str, unicode)):
            d[k] = d[k].strip()
            if not d[k]:
                del d[k]
        elif isinstance(d[k], (list, tuple)):
            new_list = []
            for item in d[k]:
                if isinstance(item, (str, unicode)):
                    if item.strip():
                        new_list.append(item.strip())
                elif item:
                    new_list.append(item)
            d[k] = new_list
            if len(d[k]) == 0:
                del d[k]
        elif isinstance(d[k], dict):
            for k2 in list(d[k].keys()):
                self.metadata_key_cleanup(d[k], k2)

    def metadata_cleanup(self):
        """ Clean metadata (delete all metadata fields without values) """
        for k in list(self.metadata.keys()):
            self.metadata_key_cleanup(self.metadata, k)

    def get_text(self):
        return self.text

    def get_references(self, reftype=None, sort=False):
        refs = self.references
        if reftype:
            refs = set([ref for ref in refs if ref.reftype == "pdf"])
        return sorted(refs) if sort else refs

    def get_references_as_dict(self, reftype=None, sort=False):
        ret = {}
        refs = self.references
        if reftype:
            refs = set([ref for ref in refs if ref.reftype == "pdf"])
        for r in sorted(refs) if sort else refs:
            if r.reftype in ret:
                ret[r.reftype].append(r.ref)
            else:
                ret[r.reftype] = [r.ref]
        return ret


class PDFMinerBackend(ReaderBackend):
    def __init__(self, pdf_stream, password="", pagenos=[], maxpages=0):  # noqa: C901
        ReaderBackend.__init__(self)
        self.pdf_stream = pdf_stream

        # Extract Metadata
        parser = PDFParser(pdf_stream)
        doc = PDFDocument(parser, password=password, caching=True)
        if doc.info:
            for k in doc.info[0]:
                v = doc.info[0][k]
                # print(repr(v), type(v))
                if isinstance(v, (bytes, str, unicode)):
                    self.metadata[k] = make_compat_str(v)
                elif isinstance(v, (psparser.PSLiteral, psparser.PSKeyword)):
                    self.metadata[k] = make_compat_str(v.name)

        # Secret Metadata
        if "Metadata" in doc.catalog:
            metadata = resolve1(doc.catalog["Metadata"]).get_data()
            # print(metadata)  # The raw XMP metadata
            # print(xmp_to_dict(metadata))
            self.metadata.update(xmp_to_dict(metadata))
            # print("---")

        # Extract Content
        text_io = BytesIO()
        rsrcmgr = PDFResourceManager(caching=True)
        converter = TextConverter(
            rsrcmgr, text_io, codec="utf-8", laparams=LAParams(), imagewriter=None
        )
        interpreter = PDFPageInterpreter(rsrcmgr, converter)

        self.metadata["Pages"] = 0
        self.curpage = 0
        for page in PDFPage.get_pages(
            self.pdf_stream,
            pagenos=pagenos,
            maxpages=maxpages,
            password=password,
            caching=True,
            check_extractable=False,
        ):
            # Read page contents
            interpreter.process_page(page)
            self.metadata["Pages"] += 1
            self.curpage += 1

            # Collect URL annotations
            # try:
            if page.annots:
                refs = self.resolve_PDFObjRef(page.annots)
                if refs:
                    if isinstance(refs, list):
                        for ref in refs:
                            if ref:
                                self.references.add(ref)
                    elif isinstance(refs, Reference):
                        self.references.add(refs)

            # except Exception as e:
            # logger.warning(str(e))

        # Remove empty metadata entries
        self.metadata_cleanup()

        # Get text from stream
        self.text = text_io.getvalue().decode("utf-8")
        text_io.close()
        converter.close()
        # print(self.text)

        # Extract URL references from text
        for url in extractor.extract_urls(self.text):
            self.references.add(Reference(url, self.curpage))

        for ref in extractor.extract_arxiv(self.text):
            self.references.add(Reference(ref, self.curpage))

        for ref in extractor.extract_doi(self.text):
            self.references.add(Reference(ref, self.curpage))

    def resolve_PDFObjRef(self, obj_ref):
        """
        Resolves PDFObjRef objects. Returns either None, a Reference object or
        a list of Reference objects.
        """
        if isinstance(obj_ref, list):
            return [self.resolve_PDFObjRef(item) for item in obj_ref]

        # print(">", obj_ref, type(obj_ref))
        if not isinstance(obj_ref, PDFObjRef):
            # print("type not of PDFObjRef")
            return None

        obj_resolved = obj_ref.resolve()
        # print("obj_resolved:", obj_resolved, type(obj_resolved))
        if isinstance(obj_resolved, bytes):
            obj_resolved = obj_resolved.decode("utf-8")

        if isinstance(obj_resolved, (str, unicode)):
            if IS_PY2:
                ref = obj_resolved.decode("utf-8")
            else:
                ref = obj_resolved
            return Reference(ref, self.curpage)

        if isinstance(obj_resolved, list):
            return [self.resolve_PDFObjRef(o) for o in obj_resolved]

        if "URI" in obj_resolved:
            if isinstance(obj_resolved["URI"], PDFObjRef):
                return self.resolve_PDFObjRef(obj_resolved["URI"])

        if "A" in obj_resolved:
            if isinstance(obj_resolved["A"], PDFObjRef):
                return self.resolve_PDFObjRef(obj_resolved["A"])

            if "URI" in obj_resolved["A"]:
                # print("->", a["A"]["URI"])
                return Reference(obj_resolved["A"]["URI"].decode("utf-8"), self.curpage)


class TextBackend(ReaderBackend):
    def __init__(self, stream):
        ReaderBackend.__init__(self)
        self.text = stream.read()

        # Extract URL references from text
        for url in extractor.extract_urls(self.text):
            self.references.add(Reference(url))

        for ref in extractor.extract_arxiv(self.text):
            self.references.add(Reference(ref))

        for ref in extractor.extract_doi(self.text):
            self.references.add(Reference(ref))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


HEADER = "\033[95m"
OKBLUE = "\033[94m"
OKGREEN = "\033[92m"
WARNING = "\033[93m"
FAIL = "\033[91m"

# Text Style
BOLD = "\033[1m"
UNDERLINE = "\033[4m"
BLINK = "\033[5m"
REVERSE = "\033[7m"

# Nothing (Standard)
ENDC = "\033[0m"


def colorprint(color, s):
    print("%s%s%s" % (color, s, ENDC))



"""
Inspired by http://stackoverflow.com/a/7257510
"""

from threading import Thread
import sys

IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    # Python 2
    from Queue import Queue
else:
    # Python 3
    from queue import Queue


class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """

    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try:
                func(*args, **kargs)
            except Exception as e:
                print(e)
            finally:
                self.tasks.task_done()


class ThreadPool:
    """ Pool of threads consuming tasks from a queue """

    def __init__(self, num_threads):
        self.tasks = Queue(num_threads)
        for _ in range(num_threads):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks.put((func, args, kargs))

    def map(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.add_task(func, args)

    def wait_completion(self):
        """ Wait for completion of all the tasks in the queue """
        self.tasks.join()


if __name__ == "__main__":
    from random import randrange
    from time import sleep

    delays = [randrange(5, 10) for i in range(100)]

    def wait_delay(d):
        print("sleeping for (%d)sec" % d)
        sleep(d)

    pool = ThreadPool(5)

    for i, d in enumerate(delays):
        pool.add_task(wait_delay, d)

    pool.wait_completion()
